@@ -15,8 +15,8 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
+ * Boston, MA 02110-1301, USA.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -246,44 +246,15 @@ gst_ffmpegscale_caps_remove_format_info (GstCaps * caps)
 {
   int i;
   GstStructure *structure;
-  GstCaps *rgbcaps;
-  GstCaps *graycaps;
 
   caps = gst_caps_copy (caps);
 
   for (i = 0; i < gst_caps_get_size (caps); i++) {
     structure = gst_caps_get_structure (caps, i);
 
-    gst_structure_set_name (structure, "video/x-raw-yuv");
     gst_structure_remove_field (structure, "format");
-    gst_structure_remove_field (structure, "endianness");
-    gst_structure_remove_field (structure, "depth");
-    gst_structure_remove_field (structure, "bpp");
-    gst_structure_remove_field (structure, "red_mask");
-    gst_structure_remove_field (structure, "green_mask");
-    gst_structure_remove_field (structure, "blue_mask");
-    gst_structure_remove_field (structure, "alpha_mask");
-    gst_structure_remove_field (structure, "palette_data");
   }
 
-  rgbcaps = gst_caps_copy (caps);
-
-  for (i = 0; i < gst_caps_get_size (rgbcaps); i++) {
-    structure = gst_caps_get_structure (rgbcaps, i);
-
-    gst_structure_set_name (structure, "video/x-raw-rgb");
-  }
-  graycaps = gst_caps_copy (caps);
-
-  for (i = 0; i < gst_caps_get_size (graycaps); i++) {
-    structure = gst_caps_get_structure (graycaps, i);
-
-    gst_structure_set_name (structure, "video/x-raw-gray");
-  }
-
-  gst_caps_append (caps, graycaps);
-  gst_caps_append (caps, rgbcaps);
-
   return caps;
 }
 
@@ -772,9 +743,9 @@ GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
     avvideoscale,
     "libav videoscaling element (" LIBAV_SOURCE ")", plugin_init,
     PACKAGE_VERSION,
-#ifdef GST_LIBAV_ENABLE_LGPL
-    "LGPL",
-#else
+#ifdef GST_LIBAV_ENABLE_GPL
     "GPL",
+#else
+    "LGPL",
 #endif
     "libav", "http://www.libav.org/")
